@@ -1,6 +1,6 @@
 ;;; vulpea-db-extract.el --- Parse and extract note data -*- lexical-binding: t; -*-
 ;;
-;; Copyright (c) 2015-2025 Boris Buliga <boris@d12frosted.io>
+;; Copyright (c) 2015-2026 Boris Buliga <boris@d12frosted.io>
 ;;
 ;; Author: Boris Buliga <boris@d12frosted.io>
 ;; Maintainer: Boris Buliga <boris@d12frosted.io>
@@ -72,7 +72,7 @@ Supports various date formats:
 - \"2025-12-08\" - ISO date
 
 Returns date string in YYYY-MM-DD format, or nil if not found."
-  (when-let ((created (cdr (assoc "CREATED" properties))))
+  (when-let* ((created (cdr (assoc "CREATED" properties))))
     (when (string-match "\\([0-9]\\{4\\}\\)-\\([0-9]\\{2\\}\\)-\\([0-9]\\{2\\}\\)" created)
       (format "%s-%s-%s"
               (match-string 1 created)
@@ -426,7 +426,7 @@ Respects `vulpea-db-index-heading-level' setting."
                        (split-string filetags-str ":" t))))
       (org-element-map ast 'headline
         (lambda (headline)
-          (when-let ((id (org-element-property :ID headline)))
+          (when-let* ((id (org-element-property :ID headline)))
             (let* ((properties (vulpea-db--extract-properties ast headline))
                    (ignored (cdr (assoc "VULPEA_IGNORE" properties)))
                    (archived (vulpea-db--archived-p headline properties filetags)))
@@ -444,8 +444,15 @@ Respects `vulpea-db-index-heading-level' setting."
                        (scheduled (org-element-property :scheduled headline))
                        (deadline (org-element-property :deadline headline))
                        (closed (org-element-property :closed headline))
-                       (meta (vulpea-db--extract-meta headline))
-                       (links (vulpea-db--extract-links headline))
+                       ;; Extract meta from the section element (first child of headline)
+                       ;; because org-element-map with NO-RECURSION 'headline won't
+                       ;; descend into a headline element itself
+                       (section (seq-find (lambda (el)
+                                            (eq (org-element-type el) 'section))
+                                          (org-element-contents headline)))
+                       (meta (when section
+                               (vulpea-db--extract-meta section)))
+                       (links (vulpea-db--extract-links headline t))
                        (outline-path (let (path
                                            (current headline))
                                        (while (setq current (org-element-property :parent current))
@@ -504,7 +511,7 @@ Respects `vulpea-db-index-heading-level' setting:
 
 Looks for property defined by `vulpea-buffer-alias-property'.
 Handles both quoted aliases (with spaces) and unquoted aliases properly."
-  (when-let ((aliases-str (cdr (assoc vulpea-buffer-alias-property properties))))
+  (when-let* ((aliases-str (cdr (assoc vulpea-buffer-alias-property properties))))
     (setq aliases-str (string-trim aliases-str))
     (let ((result nil)
           (pos 0))
@@ -557,17 +564,54 @@ Returns list of plists with :dest, :type, and :pos.
 Captures all link types: id:, roam:, file:, http:, https:,
 attachment:, elisp:, and any other `org-mode' link type.
 
-If NO-RECURSION is non-nil, stops recursion at headline boundaries.
-This is useful for file-level extraction to avoid collecting links
-from child headlines."
-  (org-element-map ast-or-node 'link
-    (lambda (link)
-      (let ((type (org-element-property :type link))
-            (path (org-element-property :path link))
-            (pos (org-element-property :begin link)))
-        (when (and type path)
-          (list :dest path :type type :pos pos))))
-    nil nil (when no-recursion 'headline)))
+If NO-RECURSION is non-nil, stops recursion at note boundaries
+\(headlines with an ID property).  Links inside non-note subtrees
+are collected as part of the current node.  This prevents links
+from child notes leaking to the parent while still capturing
+links from plain (non-note) subtrees."
+  (if no-recursion
+      (vulpea-db--extract-links-stopping-at-notes ast-or-node)
+    ;; Normal case: search everything without restrictions
+    (org-element-map ast-or-node 'link
+      (lambda (link)
+        (let ((type (org-element-property :type link))
+              (path (org-element-property :path link))
+              (pos (org-element-property :begin link)))
+          (when (and type path)
+            (list :dest path :type type :pos pos)))))))
+
+(defun vulpea-db--extract-links-stopping-at-notes (node)
+  "Extract links from NODE, stopping at note boundaries.
+
+Collects links from NODE but does not descend into child
+headlines that have an ID property (note boundaries).  Links
+inside non-note child headlines are included."
+  (let ((result nil))
+    (vulpea-db--walk-links-skipping-notes node (lambda (link) (push link result)))
+    (nreverse result)))
+
+(defun vulpea-db--walk-links-skipping-notes (node callback)
+  "Walk NODE collecting links via CALLBACK, skipping note headlines.
+
+CALLBACK is called with a plist (:dest :type :pos) for each link.
+Descends into child headlines only if they lack an ID property."
+  (dolist (child (org-element-contents node))
+    (let ((child-type (org-element-type child)))
+      (cond
+       ;; Link element: collect it
+       ((eq child-type 'link)
+        (let ((type (org-element-property :type child))
+              (path (org-element-property :path child))
+              (pos (org-element-property :begin child)))
+          (when (and type path)
+            (funcall callback (list :dest path :type type :pos pos)))))
+       ;; Headline with ID: skip (note boundary)
+       ((and (eq child-type 'headline)
+             (org-element-property :ID child))
+        nil)
+       ;; Anything else (section, paragraph, headline without ID, etc.): recurse
+       ((org-element-contents child)
+        (vulpea-db--walk-links-skipping-notes child callback))))))
 
 (defun vulpea-db--extract-meta (element)
   "Extract metadata from ELEMENT (AST or headline element).
@@ -625,7 +669,7 @@ Schema format matches `vulpea-db--schema':
    ...)
 
 Also registers the schema version in schema-registry table."
-  (when-let ((schema (vulpea-extractor-schema extractor))
+  (when-let* ((schema (vulpea-extractor-schema extractor))
              (name (vulpea-extractor-name extractor))
              (version (vulpea-extractor-version extractor)))
     (let ((db (vulpea-db)))
@@ -757,7 +801,7 @@ Returns number of notes updated (file-level + headings)."
 
       ;; Insert file-level note
       (let ((file-data (vulpea-parse-ctx-file-node ctx)))
-        (when-let ((id (plist-get file-data :id)))
+        (when-let* ((id (plist-get file-data :id)))
           (vulpea-db--insert-note-from-plist ctx path 0 0 file-data)
           (push id ids)
           (setq count (1+ count))))
@@ -770,7 +814,7 @@ Returns number of notes updated (file-level + headings)."
          (plist-get heading-data :level)
          (plist-get heading-data :pos)
          heading-data)
-        (when-let ((id (plist-get heading-data :id)))
+        (when-let* ((id (plist-get heading-data :id)))
           (push id ids))
         (setq count (1+ count)))
 
