@@ -1,13 +1,13 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "vulpea" "2.0.0"
-  "A collection of note-taking functions."
+(define-package "vulpea" "2.0.1"
+  "Note management library for Org mode."
   '((emacs   "27.2")
     (org     "9.4.4")
     (emacsql "4.3.0")
     (s       "1.12")
     (dash    "2.19"))
   :url "https://github.com/d12frosted/vulpea"
-  :commit "89ed53b762acf0c2d1985ff1501db7eac6ef412f"
-  :revdesc "v2.0.0-0-g89ed53b762ac"
+  :commit "54d5740f899e949bb492468769e1afc28d45c201"
+  :revdesc "v2.0.1-0-g54d5740f899e"
   :authors '(("Boris Buliga" . "boris@d12frosted.io"))
   :maintainers '(("Boris Buliga" . "boris@d12frosted.io")))
